// System includes
#include <stdio.h>
#include <assert.h>
#include "timer.h"

#ifndef BLOCK_SIZE
#define BLOCK_SIZE 32
#endif

int GPU = 0;

#ifdef DEBUG
	#define CUDA_CHECK() { cudaError_t error; if ((error = cudaGetLastError())!=cudaSuccess){ fprintf(stdout,"%s: %s\n", cudaGetErrorName(error),cudaGetErrorString(error));exit(EXIT_FAILURE); } }
#else
	#define CUDA_CHECK()
#endif

/**
 * Matrix multiplication (CUDA Kernel) on the device: C = A * B
 * wA is A's width and wB is B's width
 */

__global__ void copy(float *dst, float *src, int size0, int size1)
{
	int row = blockDim.y * blockIdx.y + threadIdx.y;
	int col = blockDim.x * blockIdx.x + threadIdx.x;
	if (row < size0 && col < size1)
		dst[row * size1 + col] = src[row * size1 + col];
}

/* 
 * function name: gpu_square_matrix_mult
 * 
 * source: https://github.com/lzhengchun/matrix-cuda/blob/master/matrix_cuda.cu
 */
__global__ void matrixMulCUDA(float *d_c, float *d_a, float *d_b, int n)
{
	__shared__ float tile_a[BLOCK_SIZE][BLOCK_SIZE];
	__shared__ float tile_b[BLOCK_SIZE][BLOCK_SIZE];

	int row = blockIdx.y * BLOCK_SIZE + threadIdx.y;
	int col = blockIdx.x * BLOCK_SIZE + threadIdx.x;
	float tmp = 0.0;
	int idx;

	for (int sub = 0; sub < gridDim.x; ++sub)
	{
		idx = row * n + sub * BLOCK_SIZE + threadIdx.x;
		if (idx >= n * n)
		{
			// n may not divisible by BLOCK_SIZE
			tile_a[threadIdx.y][threadIdx.x] = 0;
		}
		else
		{
			tile_a[threadIdx.y][threadIdx.x] = d_a[idx];
		}

		idx = (sub * BLOCK_SIZE + threadIdx.y) * n + col;
		if (idx >= n * n)
		{
			tile_b[threadIdx.y][threadIdx.x] = 0;
		}
		else
		{
			tile_b[threadIdx.y][threadIdx.x] = d_b[idx];
		}
		__syncthreads();

		for (int k = 0; k < BLOCK_SIZE; ++k)
		{
			tmp += tile_a[threadIdx.y][k] * tile_b[k][threadIdx.x];
		}
		__syncthreads();
	}
	if (row < n && col < n)
	{
		d_c[row * n + col] = tmp;
	}
}

/**
 * Program main
 */
int main(int argc, char **argv)
{

	// ARTURO: Simpler interface for square matrices (only one parameter)
	if (argc != 4)
	{
		fprintf(stderr, "Usage: %s <matrixSize> <n_power> <GPU>\n\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	int SIZE = atoi(argv[1]);
	int POWER = atoi(argv[2]);
	GPU = atoi(argv[3]);
	int MATRIX_SIZE = sizeof(float) * SIZE * SIZE;

	// By default, we use device 0, otherwise we override the device ID based on what is provided at the command line

	cudaSetDevice(GPU);
	CUDA_CHECK();

	// cudaError_t error;
	cudaDeviceProp deviceProp;

	Timer tTotal;
	TimerCreate(tTotal);

	// Use a larger block size for Fermi and above
	int block_size = BLOCK_SIZE;

	dim3 dimsA;
	dim3 dimsB;

	// ARTURO: Simpler interface for square matrices (only one parameter)
	dimsA.x = dimsA.y = dimsB.x = dimsB.y = SIZE;

	dim3 threads(block_size, block_size);
	dim3 grid((dimsB.x + block_size - 1) / threads.x, (dimsA.y + block_size - 1) / threads.y);

	/*Variables*/
	float *A, *C;
	cudaMallocHost(&A, MATRIX_SIZE);
	cudaMallocHost(&C, MATRIX_SIZE);
	CUDA_CHECK();

	for (int i = 0; i < SIZE * SIZE; i++)
	{
		A[i] = 2.0;
		C[i] = 0.0;
	}

	/*Device variables*/
	float *d_A, *d_C, *d_Ci;
	cudaMalloc(&d_A, MATRIX_SIZE);
	cudaMalloc(&d_C, MATRIX_SIZE);
	cudaMalloc(&d_Ci, MATRIX_SIZE);
	CUDA_CHECK();

	cudaStream_t stream1, stream2, stream3;
	cudaStreamCreate(&stream1);
	cudaStreamCreate(&stream2);
	cudaStreamCreate(&stream3);
	CUDA_CHECK();

	cudaEvent_t event1, event2, event3;
	cudaEventCreate(&event1);
	cudaEventCreate(&event2);
	cudaEventCreate(&event3);
	CUDA_CHECK();

	/*Copia al device*/
	cudaMemcpyAsync(d_A, A, MATRIX_SIZE, cudaMemcpyHostToDevice, stream1);
	cudaEventRecord(event1, stream1);
	CUDA_CHECK();

	cudaStreamWaitEvent(stream3, event1, 0);
	copy<<<grid, threads, 0, stream3>>>(d_Ci, d_A, SIZE, SIZE);
	cudaEventRecord(event3, stream3);
	CUDA_CHECK();

	/* Start timer */
	cudaDeviceSynchronize();
	CUDA_CHECK();
	TimerStart(tTotal);

	for (int i = 2; i < POWER; i++)
	{
		matrixMulCUDA<<<grid, threads, 0, stream3>>>(d_C, d_Ci, d_A, SIZE);
		copy<<<grid, threads, 0, stream3>>>(d_Ci, d_C, SIZE, SIZE);
		cudaEventRecord(event3, stream3);

		cudaStreamWaitEvent(stream2, event3, 0);
		cudaMemcpyAsync(C, d_C, MATRIX_SIZE, cudaMemcpyDeviceToHost, stream2);
	}
	cudaDeviceSynchronize();
	CUDA_CHECK();

	/* Stop timer */
	TimerStop(tTotal);

	cudaEventDestroy(event1);
	cudaEventDestroy(event2);
	cudaEventDestroy(event3);
	cudaStreamDestroy(stream1);
	cudaStreamDestroy(stream2);
	cudaStreamDestroy(stream3);
	cudaFree(d_A);
	cudaFree(d_C);
	cudaFree(d_Ci);
	CUDA_CHECK();

	printf("\nVERSION A_SINCRONA************\n");
	printf("SIZE %d\n", SIZE);
	printf("Clock Main %.8lf <--\n", TimerGetTime(tTotal));
	TimerDestroy(tTotal);

	// 11. Calculate NORM
	resultado = 0, suma = 0;
	for (int i = 0; i < SIZE; i++)
	{
		for (int j = 0; j < SIZE; j++)
		{
			suma += pow(C[i * SIZE + j], 2);
		}
	}

	printf("\n ----------------------- NORM ----------------------- \n");
	printf("\n Acumulated sum: %lf", suma);

	resultado = sqrt(suma);
	printf("\n Result: %lf \n", resultado);

	printf("\n ---------------------------------------------------- \n");

	cudaFreeHost(A);
	cudaFreeHost(C);
	CUDA_CHECK();

	return (0);
}
